@@ -41,14 +41,6 @@ host_logger.setLevel(logging.INFO)
 NUM_RETRIES = 3
 
 logger = logging.getLogger(__name__)
-
-def _setup_logger(_logger):
-    """Setup default logger"""
-    _handler = logging.StreamHandler()
-    log_format = logging.Formatter('%(name)s - %(asctime)s - %(levelname)s - %(message)s')
-    _handler.setFormatter(log_format)
-    _logger.addHandler(handler)
-    _logger.setLevel(logging.DEBUG)
     
 class UnknownHostException(Exception):
     """Raised when a host is unknown (dns failure)"""
@@ -122,10 +114,7 @@ class SSHClient(object):
                             'hostname' in host_config
                             else host)
         _user = host_config['user'] if 'user' in host_config else None
-        if user:
-            user = user
-        else:
-            user = _user
+        user = user if user else _user
         client = paramiko.SSHClient()
         client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
         self.forward_ssh_agent = forward_ssh_agent
@@ -139,8 +128,8 @@ class SSHClient(object):
         self.proxy_command = paramiko.ProxyCommand(host_config['proxycommand']) if 'proxycommand' in \
                              host_config else None
         if self.proxy_command:
-            logger.debug("Proxy configured for destination host %s - ProxyCommand: '%s'" % (
-                self.host, " ".join(self.proxy_command.cmd),))
+            logger.debug("Proxy configured for destination host %s - ProxyCommand: '%s'",
+                         self.host, " ".join(self.proxy_command.cmd),)
         if _agent:
             self.client._agent = _agent
         self._connect()
@@ -153,26 +142,26 @@ class SSHClient(object):
                                 pkey=self.pkey,
                                 sock=self.proxy_command)
         except socket.gaierror, e:
-            logger.error("Could not resolve host '%s'", self.host,)
+            logger.error("Could not resolve host '%s' - retry %s/%s",
+                         self.host, retries, NUM_RETRIES)
             while retries < NUM_RETRIES:
                 gevent.sleep(5)
                 return self._connect(retries=retries+1)
             raise UnknownHostException("%s - %s" % (str(e.args[1]),
                                                     self.host,))
         except socket.error, e:
-            logger.error("Error connecting to host '%s:%s'" % (self.host,
-                                                               self.port,))
+            logger.error("Error connecting to host '%s:%s' - retry %s/%s",
+                         self.host, self.port, retries, NUM_RETRIES)
             while retries < NUM_RETRIES:
                 gevent.sleep(5)
                 return self._connect(retries=retries+1)
-            raise ConnectionErrorException("%s for host '%s:%s'" % (str(e.args[1]),
-                                                                    self.host,
-                                                                    self.port,))
+            raise ConnectionErrorException("%s for host '%s:%s' - retry %s/%s",
+                                           str(e.args[1]), self.host, self.port,
+                                           retries, NUM_RETRIES,)
         except paramiko.AuthenticationException, e:
             raise AuthenticationException(e)
         except paramiko.ProxyCommandFailure, e:
-            logger.error("Error executing ProxyCommand - %s" % (
-                e.message,))
+            logger.error("Error executing ProxyCommand - %s", e.message,)
             raise ProxyCommandException(e.message)
 
     def exec_command(self, command, sudo=False, **kwargs):
@@ -188,7 +177,7 @@ class SSHClient(object):
         :type sudo: bool
         :param kwargs: (Optional) Keyword arguments to be passed to remote \
         command
-        :type kwargs: dict                
+        :type kwargs: dict
         :rtype: Tuple of `(channel, hostname, stdout, stderr)`. \
         Channel is the remote SSH channel, needed to ensure all of stdout has \
         been got, hostname is remote hostname the copy is to, stdout and \
@@ -215,7 +204,7 @@ class SSHClient(object):
     def _make_sftp(self):
         """Make SFTP client from open transport"""
         transport = self.client.get_transport()
-        channel = transport.open_session()
+        transport.open_session()
         return paramiko.SFTPClient.from_transport(transport)
 
     def mkdir(self, sftp, directory):
@@ -246,7 +235,6 @@ class SSHClient(object):
         """
         sftp = self._make_sftp()
         destination = remote_file.split(os.path.sep)
-        filename = destination[0] if len(destination) == 1 else destination[-1]
         remote_file = os.path.sep.join(destination)
         destination = destination[:-1]
         for directory in destination:
@@ -289,6 +277,10 @@ class ParallelSSHClient(object):
         :type port: int
         :param pkey: (Optional) Client's private key to be used to connect with
         :type pkey: :mod:`paramiko.PKey`
+        :param forward_ssh_agent: (Optional) Turn on SSH agent forwarding - \
+        equivalent to `ssh -A` from the `ssh` command line utility. \
+        Defaults to True if not set.
+        :type forward_ssh_agent: bool
         :param pool_size: (Optional) Greenlet pool size. Controls on how many\
         hosts to execute tasks in parallel. Defaults to 10
         :type pool_size: int
@@ -438,23 +430,48 @@ class ParallelSSHClient(object):
         :rtype: Dictionary containing ``{host: {'exit_code': exit code}}`` entry \
         for example ``{'myhost1': {'exit_code': 0}}``
         :rtype: With ``return_buffers=True``: ``{'myhost1': {'exit_code': 0,
+                                                             'channel' : None or SSH channel of command if command is still executing,
          						     'stdout' : <iterable>,
                                                              'stderr' : <iterable>,}}``
         """
         channel, host, _stdout, _stderr = greenlet.get()
-        stdout = (line.strip() for line in _stdout)
-        stderr = (line.strip() for line in _stderr)
-        channel.close()
-        if not return_buffers:
+        stdout = self._read_output_buffer(_stdout)
+        stderr = self._read_output_buffer(_stderr)
+        if channel.exit_status_ready():
+            channel.close()
+        else:
+            logger.debug("Command still executing on get_stdout call - not closing channel and returning None as exit code.")
+            # If channel is not closed we cannot get full stdout/stderr so must return buffers
+            return_buffers = True
+        # Channel must be closed or reading stdout/stderr will block forever
+        if not return_buffers and channel.closed:
             for line in stdout:
                 host_logger.info("[%s]\t%s", host, line,)
             for line in stderr:
                 host_logger.info("[%s] [err] %s", host, line,)
             return {host: {'exit_code': channel.recv_exit_status(),}}
-        return {host: {'exit_code': channel.recv_exit_status(),
+        gevent.sleep(.2)
+        return {host: {'exit_code': channel.recv_exit_status() if channel.exit_status_ready() else None,
+                       'channel' : channel if not channel.closed else None,
                        'stdout' : stdout,
                        'stderr' : stderr, }}
 
+    def wait_for_exit_status(self, channel):
+        """Block and wait for exit status on channel.
+        WARNING - this will block forever if the command executed never exits
+        :rtype: int - Exit code of command executed"""
+        while not channel.exit_status_ready():
+            gevent.sleep(1)
+        channel.close()
+        return channel.recv_exit_status()
+
+    def _read_output_buffer(self, output_buffer):
+        """Read from output buffers,
+        allowing coroutines to execute in between reading"""
+        for line in output_buffer:
+            gevent.sleep(1)
+            yield line.strip()
+
     def copy_file(self, local_file, remote_file):
         """Copy local file to remote file in parallel
         
@@ -486,24 +503,3 @@ class ParallelSSHClient(object):
                                                 port=self.port, pkey=self.pkey,
                                                 forward_ssh_agent=self.forward_ssh_agent)
         return self.host_clients[host].copy_file(local_file, remote_file)
-
-    
-def test():
-    client = SSHClient('localhost')
-    channel, host, stdout, stderr = client.exec_command('ls -ltrh')
-    for line in stdout:
-        print line.strip()
-    client.copy_file('../test', 'test_dir/test')
-
-def test_parallel():
-    client = ParallelSSHClient(['localhost'])
-    cmds = client.exec_command('ls -ltrh')
-    output = [client.get_stdout(cmd, return_buffers=True) for cmd in cmds]
-    print output
-    cmds = client.copy_file('../test', 'test_dir/test')
-    client.pool.join()
-
-if __name__ == "__main__":
-    _setup_logger(logger)
-    test()
-    test_parallel()
